@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # -*- coding: utf-8 -*-
 
 # Copyright: (c) 2012, Flowroute LLC
@@ -21,7 +20,7 @@ version_added: "0.0.2"
 options:
   name:
     description:
-      - A list of package names, like C(foo), or package specifier with version, like C(foo=1.0).
+      - A list of package names, like C(foo), or package specifier with version, like C(foo=1.0) or C(foo>=1.0).
         Name wildcards (fnmatch) like C(apt*) and version wildcards like C(foo=1.0*) are also supported.
     aliases: [ package, pkg ]
     type: list
@@ -83,6 +82,14 @@ options:
          Please also see C(man apt-get) for more information.'
     type: bool
     default: 'no'
+  clean:
+    description:
+      - Run the equivalent of C(apt-get clean) to clear out the local repository of retrieved package files. It removes everything but
+        the lock file from /var/cache/apt/archives/ and /var/cache/apt/archives/partial/.
+      - Can be run as part of the package installation (clean runs before install) or as a separate step.
+    type: bool
+    default: 'no'
+    version_added: "2.13"
   allow_unauthenticated:
     description:
       - Ignore if packages cannot be authenticated. This is useful for bootstrapping environments that manage their own apt-key setup.
@@ -101,6 +108,12 @@ options:
     type: bool
     default: 'no'
     version_added: "2.12"
+  allow_change_held_packages:
+    description:
+      - Allows changing the version of a package which is on the apt hold list
+    type: bool
+    default: 'no'
+    version_added: '2.13'
   upgrade:
     description:
       - If yes or safe, performs an aptitude safe-upgrade.
@@ -170,7 +183,7 @@ options:
     version_added: "2.4"
   lock_timeout:
     description:
-      - How many seconds will this action wait to aquire a lock on the apt db.
+      - How many seconds will this action wait to acquire a lock on the apt db.
       - Sometimes there is a transitory lock and this will retry at least until timeout is hit.
     type: int
     default: 60
@@ -196,120 +209,114 @@ notes:
    - The apt-get commandline supports implicit regex matches here but we do not because it can let typos through easier
      (If you typo C(foo) as C(fo) apt-get would install packages that have "fo" in their name with a warning and a prompt for the user.
      Since we don't have warnings and prompts before installing we disallow this.Use an explicit fnmatch pattern if you want wildcarding)
-   - When used with a `loop:` each package will be processed individually, it is much more efficient to pass the list directly to the `name` option.
+   - When used with a C(loop:) each package will be processed individually, it is much more efficient to pass the list directly to the I(name) option.
+   - When C(default_release) is used, an implicit priority of 990 is used. This is the same behavior as C(apt-get -t).
+   - When an exact version is specified, an implicit priority of 1001 is used.
 '''
 
 EXAMPLES = '''
 - name: Install apache httpd  (state=present is optional)
-  apt:
+  ansible.builtin.apt:
     name: apache2
     state: present
 
 - name: Update repositories cache and install "foo" package
-  apt:
+  ansible.builtin.apt:
     name: foo
     update_cache: yes
 
 - name: Remove "foo" package
-  apt:
+  ansible.builtin.apt:
     name: foo
     state: absent
 
 - name: Install the package "foo"
-  apt:
+  ansible.builtin.apt:
     name: foo
 
 - name: Install a list of packages
-  apt:
+  ansible.builtin.apt:
     pkg:
     - foo
     - foo-tools
 
 - name: Install the version '1.00' of package "foo"
-  apt:
+  ansible.builtin.apt:
     name: foo=1.00
 
 - name: Update the repository cache and update package "nginx" to latest version using default release squeeze-backport
-  apt:
+  ansible.builtin.apt:
     name: nginx
     state: latest
     default_release: squeeze-backports
     update_cache: yes
 
 - name: Install the version '1.18.0' of package "nginx" and allow potential downgrades
-  apt:
+  ansible.builtin.apt:
     name: nginx=1.18.0
     state: present
     allow_downgrade: yes
 
 - name: Install zfsutils-linux with ensuring conflicted packages (e.g. zfs-fuse) will not be removed.
-  apt:
+  ansible.builtin.apt:
     name: zfsutils-linux
     state: latest
     fail_on_autoremove: yes
 
 - name: Install latest version of "openjdk-6-jdk" ignoring "install-recommends"
-  apt:
+  ansible.builtin.apt:
     name: openjdk-6-jdk
     state: latest
     install_recommends: no
 
 - name: Update all packages to their latest version
-  apt:
+  ansible.builtin.apt:
     name: "*"
     state: latest
 
 - name: Upgrade the OS (apt-get dist-upgrade)
-  apt:
+  ansible.builtin.apt:
     upgrade: dist
 
 - name: Run the equivalent of "apt-get update" as a separate step
-  apt:
+  ansible.builtin.apt:
     update_cache: yes
 
 - name: Only run "update_cache=yes" if the last one is more than 3600 seconds ago
-  apt:
+  ansible.builtin.apt:
     update_cache: yes
     cache_valid_time: 3600
 
 - name: Pass options to dpkg on run
-  apt:
+  ansible.builtin.apt:
     upgrade: dist
     update_cache: yes
     dpkg_options: 'force-confold,force-confdef'
 
 - name: Install a .deb package
-  apt:
+  ansible.builtin.apt:
     deb: /tmp/mypackage.deb
 
 - name: Install the build dependencies for package "foo"
-  apt:
+  ansible.builtin.apt:
     pkg: foo
     state: build-dep
 
 - name: Install a .deb package from the internet
-  apt:
+  ansible.builtin.apt:
     deb: https://example.com/python-ppq_0.1-1_all.deb
 
 - name: Remove useless packages from the cache
-  apt:
+  ansible.builtin.apt:
     autoclean: yes
 
 - name: Remove dependencies that are no longer required
-  apt:
+  ansible.builtin.apt:
     autoremove: yes
 
-# Sometimes apt tasks fail because apt is locked by an autoupdate or by a race condition on a thread.
-# To check for a lock file before executing, and keep trying until the lock file is released:
-- name: Install packages only when the apt process is not locked
+- name: Run the equivalent of "apt-get clean" as a separate step
   apt:
-    name: foo
-    state: present
-  register: apt_action
-  retries: 100
-  until: apt_action is success or ('Failed to lock apt for exclusive operation' not in apt_action.msg and '/var/lib/dpkg/lock' not in apt_action.msg)
-
-
+    clean: yes
 '''
 
 RETURN = '''
@@ -327,7 +334,12 @@ stdout:
     description: output from apt
     returned: success, when needed
     type: str
-    sample: "Reading package lists...\nBuilding dependency tree...\nReading state information...\nThe following extra packages will be installed:\n  apache2-bin ..."
+    sample: |-
+        Reading package lists...
+        Building dependency tree...
+        Reading state information...
+        The following extra packages will be installed:
+          apache2-bin ...
 stderr:
     description: error output from apt
     returned: success, when needed
@@ -353,8 +365,8 @@ import time
 from ansible.module_utils.basic import AnsibleModule
 from ansible.module_utils.common.locale import get_best_parsable_locale
 from ansible.module_utils.common.respawn import has_respawned, probe_interpreters_for_module, respawn_module
-from ansible.module_utils._text import to_bytes, to_native
-from ansible.module_utils.six import PY3
+from ansible.module_utils._text import to_native, to_text
+from ansible.module_utils.six import PY3, string_types
 from ansible.module_utils.urls import fetch_file
 
 DPKG_OPTIONS = 'force-confdef,force-confold'
@@ -371,7 +383,6 @@ CLEAN_OP_CHANGED_STR = dict(
     autoclean='Del ',
 )
 
-apt = apt_pkg = None  # keep pylint happy by declaring unconditionally
 
 HAS_PYTHON_APT = False
 try:
@@ -380,7 +391,7 @@ try:
     import apt_pkg
     HAS_PYTHON_APT = True
 except ImportError:
-    pass
+    apt = apt_pkg = None
 
 
 class PolicyRcD(object):
@@ -461,24 +472,10 @@ class PolicyRcD(object):
 
 
 def package_split(pkgspec):
-    parts = pkgspec.split('=', 1)
-    version = None
+    parts = re.split(r'(>?=)', pkgspec, 1)
     if len(parts) > 1:
-        version = parts[1]
-    return parts[0], version
-
-
-def package_versions(pkgname, pkg, pkg_cache):
-    try:
-        versions = set(p.version for p in pkg.versions)
-    except AttributeError:
-        # assume older version of python-apt is installed
-        # apt.package.Package#versions require python-apt >= 0.7.9.
-        pkg_cache_list = (p for p in pkg_cache.Packages if p.Name == pkgname)
-        pkg_versions = (p.VersionList for p in pkg_cache_list)
-        versions = set(p.VerStr for p in itertools.chain(*pkg_versions))
-
-    return versions
+        return parts
+    return parts[0], None, None
 
 
 def package_version_compare(version, other_version):
@@ -488,7 +485,39 @@ def package_version_compare(version, other_version):
         return apt_pkg.VersionCompare(version, other_version)
 
 
-def package_status(m, pkgname, version, cache, state):
+def package_best_match(pkgname, version_cmp, version, release, cache):
+    policy = apt_pkg.Policy(cache)
+
+    policy.read_pinfile(apt_pkg.config.find_file("Dir::Etc::preferences"))
+    policy.read_pindir(apt_pkg.config.find_file("Dir::Etc::preferencesparts"))
+
+    if release:
+        # 990 is the priority used in `apt-get -t`
+        policy.create_pin('Release', pkgname, release, 990)
+    if version_cmp == "=":
+        # Installing a specific version from command line overrides all pinning
+        # We don't mimmic this exactly, but instead set a priority which is higher than all APT built-in pin priorities.
+        policy.create_pin('Version', pkgname, version, 1001)
+    pkg = cache[pkgname]
+    pkgver = policy.get_candidate_ver(pkg)
+    if not pkgver:
+        return None
+    if version_cmp == "=" and not fnmatch.fnmatch(pkgver.ver_str, version):
+        # Even though we put in a pin policy, it can be ignored if there is no
+        # possible candidate.
+        return None
+    return pkgver.ver_str
+
+
+def package_status(m, pkgname, version_cmp, version, default_release, cache, state):
+    """
+    :return: A tuple of (installed, installed_version, version_installable, has_files). *installed* indicates whether
+    the package (regardless of version) is installed. *installed_version* indicates whether the installed package
+    matches the provided version criteria. *version_installable* provides the latest matching version that can be
+    installed. In the case of virtual packages where we can't determine an applicable match, True is returned.
+    *has_files* indicates whether the package has files on the filesystem (even if not installed, meaning a purge is
+    required).
+    """
     try:
         # get the package from the cache, as well as the
         # low-level apt_pkg.Package object which contains
@@ -501,23 +530,27 @@ def package_status(m, pkgname, version, cache, state):
             try:
                 provided_packages = cache.get_providing_packages(pkgname)
                 if provided_packages:
-                    is_installed = False
-                    upgradable = False
-                    version_ok = False
-                    # when virtual package providing only one package, look up status of target package
+                    # When this is a virtual package satisfied by only
+                    # one installed package, return the status of the target
+                    # package to avoid requesting re-install
                     if cache.is_virtual_package(pkgname) and len(provided_packages) == 1:
                         package = provided_packages[0]
-                        installed, version_ok, upgradable, has_files = package_status(m, package.name, version, cache, state='install')
+                        installed, installed_version, version_installable, has_files = \
+                            package_status(m, package.name, version_cmp, version, default_release, cache, state='install')
                         if installed:
-                            is_installed = True
-                    return is_installed, version_ok, upgradable, False
+                            return installed, installed_version, version_installable, has_files
+
+                    # Otherwise return nothing so apt will sort out
+                    # what package to satisfy this with
+                    return False, False, True, False
+
                 m.fail_json(msg="No package matching '%s' is available" % pkgname)
             except AttributeError:
                 # python-apt version too old to detect virtual packages
-                # mark as upgradable and let apt-get install deal with it
+                # mark as not installed and let apt-get install deal with it
                 return False, False, True, False
         else:
-            return False, False, False, False
+            return False, False, None, False
     try:
         has_files = len(pkg.installed_files) > 0
     except UnicodeDecodeError:
@@ -535,36 +568,32 @@ def package_status(m, pkgname, version, cache, state):
             # assume older version of python-apt is installed
             package_is_installed = pkg.isInstalled
 
-    version_is_installed = package_is_installed
-    if version:
-        versions = package_versions(pkgname, pkg, cache._cache)
-        avail_upgrades = fnmatch.filter(versions, version)
-
-        if package_is_installed:
-            try:
-                installed_version = pkg.installed.version
-            except AttributeError:
-                installed_version = pkg.installedVersion
+    version_best = package_best_match(pkgname, version_cmp, version, default_release, cache._cache)
+    version_is_installed = False
+    version_installable = None
+    if package_is_installed:
+        try:
+            installed_version = pkg.installed.version
+        except AttributeError:
+            installed_version = pkg.installedVersion
 
+        if version_cmp == "=":
             # check if the version is matched as well
             version_is_installed = fnmatch.fnmatch(installed_version, version)
-
-            # Only claim the package is upgradable if a candidate matches the version
-            package_is_upgradable = False
-            for candidate in avail_upgrades:
-                if package_version_compare(candidate, installed_version) > 0:
-                    package_is_upgradable = True
-                    break
+            if version_best and installed_version != version_best and fnmatch.fnmatch(version_best, version):
+                version_installable = version_best
+        elif version_cmp == ">=":
+            version_is_installed = apt_pkg.version_compare(installed_version, version) >= 0
+            if version_best and installed_version != version_best and apt_pkg.version_compare(version_best, version) >= 0:
+                version_installable = version_best
         else:
-            package_is_upgradable = bool(avail_upgrades)
+            version_is_installed = True
+            if version_best and installed_version != version_best:
+                version_installable = version_best
     else:
-        try:
-            package_is_upgradable = pkg.is_upgradable
-        except AttributeError:
-            # assume older version of python-apt is installed
-            package_is_upgradable = pkg.isUpgradable
+        version_installable = version_best
 
-    return package_is_installed, version_is_installed, package_is_upgradable, has_files
+    return package_is_installed, version_is_installed, version_installable, has_files
 
 
 def expand_dpkg_options(dpkg_options_compressed):
@@ -588,7 +617,11 @@ def expand_pkgspec_from_fnmatches(m, pkgspec, cache):
     new_pkgspec = []
     if pkgspec:
         for pkgspec_pattern in pkgspec:
-            pkgname_pattern, version = package_split(pkgspec_pattern)
+
+            if not isinstance(pkgspec_pattern, string_types):
+                m.fail_json(msg="Invalid type for package name, expected string but got %s" % type(pkgspec_pattern))
+
+            pkgname_pattern, version_cmp, version = package_split(pkgspec_pattern)
 
             # note that none of these chars is allowed in a (debian) pkgname
             if frozenset('*?[]!').intersection(pkgname_pattern):
@@ -597,20 +630,20 @@ def expand_pkgspec_from_fnmatches(m, pkgspec, cache):
                 if ":" not in pkgname_pattern:
                     # Filter the multiarch packages from the cache only once
                     try:
-                        pkg_name_cache = _non_multiarch
+                        pkg_name_cache = _non_multiarch  # pylint: disable=used-before-assignment
                     except NameError:
                         pkg_name_cache = _non_multiarch = [pkg.name for pkg in cache if ':' not in pkg.name]  # noqa: F841
                 else:
                     # Create a cache of pkg_names including multiarch only once
                     try:
-                        pkg_name_cache = _all_pkg_names
+                        pkg_name_cache = _all_pkg_names  # pylint: disable=used-before-assignment
                     except NameError:
                         pkg_name_cache = _all_pkg_names = [pkg.name for pkg in cache]  # noqa: F841
 
                 matches = fnmatch.filter(pkg_name_cache, pkgname_pattern)
 
                 if not matches:
-                    m.fail_json(msg="No package(s) matching '%s' available" % str(pkgname_pattern))
+                    m.fail_json(msg="No package(s) matching '%s' available" % to_text(pkgname_pattern))
                 else:
                     new_pkgspec.extend(matches)
             else:
@@ -667,7 +700,7 @@ def install(m, pkgspec, cache, upgrade=False, default_release=None,
             install_recommends=None, force=False,
             dpkg_options=expand_dpkg_options(DPKG_OPTIONS),
             build_dep=False, fixed=False, autoremove=False, fail_on_autoremove=False, only_upgrade=False,
-            allow_unauthenticated=False, allow_downgrade=False):
+            allow_unauthenticated=False, allow_downgrade=False, allow_change_held_packages=False):
     pkg_list = []
     packages = ""
     pkgspec = expand_pkgspec_from_fnmatches(m, pkgspec, cache)
@@ -678,19 +711,35 @@ def install(m, pkgspec, cache, upgrade=False, default_release=None,
             pkg_list.append("'%s'" % package)
             continue
 
-        name, version = package_split(package)
+        name, version_cmp, version = package_split(package)
         package_names.append(name)
-        installed, installed_version, upgradable, has_files = package_status(m, name, version, cache, state='install')
-        if (not installed and not only_upgrade) or (installed and not installed_version) or (upgrade and upgradable):
-            pkg_list.append("'%s'" % package)
-        if installed_version and upgradable and version:
+        installed, installed_version, version_installable, has_files = package_status(m, name, version_cmp, version, default_release, cache, state='install')
+
+        if not installed and only_upgrade:
+            # only_upgrade upgrades packages that are already installed
+            # since this package is not installed, skip it
+            continue
+
+        if not installed_version and not version_installable:
+            status = False
+            data = dict(msg="no available installation candidate for %s" % package)
+            return (status, data)
+
+        if version_installable and ((not installed and not only_upgrade) or upgrade or not installed_version):
+            if version_installable is not True:
+                pkg_list.append("'%s=%s'" % (name, version_installable))
+            elif version:
+                pkg_list.append("'%s=%s'" % (name, version))
+            else:
+                pkg_list.append("'%s'" % name)
+        elif installed_version and version_installable and version_cmp == "=":
             # This happens when the package is installed, a newer version is
             # available, and the version is a wildcard that matches both
             #
-            # We do not apply the upgrade flag because we cannot specify both
-            # a version and state=latest.  (This behaviour mirrors how apt
-            # treats a version with wildcard in the package)
-            pkg_list.append("'%s'" % package)
+            # This is legacy behavior, and isn't documented (in fact it does
+            # things documentations says it shouldn't). It should not be relied
+            # upon.
+            pkg_list.append("'%s=%s'" % (name, version))
     packages = ' '.join(pkg_list)
 
     if packages:
@@ -745,6 +794,9 @@ def install(m, pkgspec, cache, upgrade=False, default_release=None,
         if allow_downgrade:
             cmd += " --allow-downgrades"
 
+        if allow_change_held_packages:
+            cmd += " --allow-change-held-packages"
+
         with PolicyRcD(m):
             rc, out, err = m.run_command(cmd)
 
@@ -766,7 +818,7 @@ def install(m, pkgspec, cache, upgrade=False, default_release=None,
         status = True
         data = dict(changed=False)
 
-    if not build_dep:
+    if not build_dep and not m.check_mode:
         mark_installed_manually(m, package_names)
 
     return (status, data)
@@ -781,16 +833,22 @@ def get_field_of_deb(m, deb_file, field="Version"):
     return to_native(stdout).strip('\n')
 
 
-def install_deb(m, debs, cache, force, fail_on_autoremove, install_recommends, allow_unauthenticated, allow_downgrade, dpkg_options):
+def install_deb(
+        m, debs, cache, force, fail_on_autoremove, install_recommends,
+        allow_unauthenticated,
+        allow_downgrade,
+        allow_change_held_packages,
+        dpkg_options,
+):
     changed = False
     deps_to_install = []
     pkgs_to_install = []
     for deb_file in debs.split(','):
         try:
-            pkg = apt.debfile.DebPackage(deb_file)
+            pkg = apt.debfile.DebPackage(deb_file, cache=apt.Cache())
             pkg_name = get_field_of_deb(m, deb_file, "Package")
             pkg_version = get_field_of_deb(m, deb_file, "Version")
-            if len(apt_pkg.get_architectures()) > 1:
+            if hasattr(apt_pkg, 'get_architectures') and len(apt_pkg.get_architectures()) > 1:
                 pkg_arch = get_field_of_deb(m, deb_file, "Architecture")
                 pkg_key = "%s:%s" % (pkg_name, pkg_arch)
             else:
@@ -829,6 +887,7 @@ def install_deb(m, debs, cache, force, fail_on_autoremove, install_recommends, a
                                      fail_on_autoremove=fail_on_autoremove,
                                      allow_unauthenticated=allow_unauthenticated,
                                      allow_downgrade=allow_downgrade,
+                                     allow_change_held_packages=allow_change_held_packages,
                                      dpkg_options=expand_dpkg_options(dpkg_options))
         if not success:
             m.fail_json(**retvals)
@@ -874,8 +933,8 @@ def remove(m, pkgspec, cache, purge=False, force=False,
     pkg_list = []
     pkgspec = expand_pkgspec_from_fnmatches(m, pkgspec, cache)
     for package in pkgspec:
-        name, version = package_split(package)
-        installed, installed_version, upgradable, has_files = package_status(m, name, version, cache, state='remove')
+        name, version_cmp, version = package_split(package)
+        installed, installed_version, upgradable, has_files = package_status(m, name, version_cmp, version, None, cache, state='remove')
         if installed_version or (has_files and purge):
             pkg_list.append("'%s'" % package)
     packages = ' '.join(pkg_list)
@@ -955,6 +1014,19 @@ def cleanup(m, purge=False, force=False, operation=None,
     m.exit_json(changed=changed, stdout=out, stderr=err, diff=diff)
 
 
+def aptclean(m):
+    clean_rc, clean_out, clean_err = m.run_command(['apt-get', 'clean'])
+    if m._diff:
+        clean_diff = parse_diff(clean_out)
+    else:
+        clean_diff = {}
+    if clean_rc:
+        m.fail_json(msg="apt-get clean failed", stdout=clean_out, rc=clean_rc)
+    if clean_err:
+        m.fail_json(msg="apt-get clean failed: %s" % clean_err, stdout=clean_out, rc=clean_rc)
+    return clean_out, clean_err
+
+
 def upgrade(m, mode="yes", force=False, default_release=None,
             use_apt_get=False,
             dpkg_options=expand_dpkg_options(DPKG_OPTIONS), autoremove=False, fail_on_autoremove=False,
@@ -1116,8 +1188,10 @@ def main():
             policy_rc_d=dict(type='int', default=None),
             only_upgrade=dict(type='bool', default=False),
             force_apt_get=dict(type='bool', default=False),
+            clean=dict(type='bool', default=False),
             allow_unauthenticated=dict(type='bool', default=False, aliases=['allow-unauthenticated']),
             allow_downgrade=dict(type='bool', default=False, aliases=['allow-downgrade', 'allow_downgrades', 'allow-downgrades']),
+            allow_change_held_packages=dict(type='bool', default=False),
             lock_timeout=dict(type='int', default=60),
         ),
         mutually_exclusive=[['deb', 'package', 'upgrade']],
@@ -1202,6 +1276,18 @@ def main():
 
     p = module.params
 
+    if p['clean'] is True:
+        aptclean_stdout, aptclean_stderr = aptclean(module)
+        # If there is nothing else to do exit. This will set state as
+        #  changed based on if the cache was updated.
+        if not p['package'] and not p['upgrade'] and not p['deb']:
+            module.exit_json(
+                changed=True,
+                msg=aptclean_stdout,
+                stdout=aptclean_stdout,
+                stderr=aptclean_stderr
+            )
+
     if p['upgrade'] == 'no':
         p['upgrade'] = None
 
@@ -1215,6 +1301,7 @@ def main():
     install_recommends = p['install_recommends']
     allow_unauthenticated = p['allow_unauthenticated']
     allow_downgrade = p['allow_downgrade']
+    allow_change_held_packages = p['allow_change_held_packages']
     dpkg_options = expand_dpkg_options(p['dpkg_options'])
     autoremove = p['autoremove']
     fail_on_autoremove = p['fail_on_autoremove']
@@ -1254,7 +1341,8 @@ def main():
 
                     for retry in range(update_cache_retries):
                         try:
-                            cache.update()
+                            if not module.check_mode:
+                                cache.update()
                             break
                         except apt.cache.FetchFailedException as e:
                             err = to_native(e)
@@ -1269,7 +1357,7 @@ def main():
 
                     cache.open(progress=None)
                     mtimestamp, post_cache_update_time = get_updated_cache_time()
-                    if updated_cache_time != post_cache_update_time:
+                    if module.check_mode or updated_cache_time != post_cache_update_time:
                         updated_cache = True
                     updated_cache_time = post_cache_update_time
 
@@ -1306,6 +1394,7 @@ def main():
                 install_deb(module, p['deb'], cache,
                             install_recommends=install_recommends,
                             allow_unauthenticated=allow_unauthenticated,
+                            allow_change_held_packages=allow_change_held_packages,
                             allow_downgrade=allow_downgrade,
                             force=force_yes, fail_on_autoremove=fail_on_autoremove, dpkg_options=p['dpkg_options'])
 
@@ -1334,8 +1423,6 @@ def main():
                 for package in packages:
                     if package.count('=') > 1:
                         module.fail_json(msg="invalid package spec: %s" % package)
-                    if latest and '=' in package:
-                        module.fail_json(msg='version number inconsistent with state=latest: %s' % package)
 
             if not packages:
                 if autoclean:
@@ -1369,7 +1456,8 @@ def main():
                     fail_on_autoremove=fail_on_autoremove,
                     only_upgrade=p['only_upgrade'],
                     allow_unauthenticated=allow_unauthenticated,
-                    allow_downgrade=allow_downgrade
+                    allow_downgrade=allow_downgrade,
+                    allow_change_held_packages=allow_change_held_packages,
                 )
 
                 # Store if the cache has been updated
